%-------------------------------------------------------------------------
% FILE: ekf.m
% AUTH: B. Green
% DATE: 07/25/06
% DESC: Implements an Extended Kalman Filter to eliminate the noise in a 
%       position measurement (simulated) of a moving vehicle.
%-------------------------------------------------------------------------

clear all; clc;

N = 10;

%---------------------SIMULATE NOISY MEASUREMENT--------------------------
z = zeros(3,N);
z(1,:) = [0, 2, 4, 10, 11, 10, 10, 10, 9, 8];
z(2,:) = [0, 1, 3, 1, 5, 9, 12, 14, 16, 18];
z(3,:) = [0, 0.25, 0.65, 1.0, 1.05, 1.4, 1.8, 2.0, 2.3, 2.5];
deltaT = 1;     % time between samples
%-------------------------------------------------------------------------

%---------------------STATE BASED ON PROCESS MODEL------------------------
x = zeros(1,N);
y = zeros(1,N);
theta = zeros(1,N);

L=1;                    % distance between rear axle and steering axle
vel = 3 + zeros(1,N);     % velocity input (V=3)
phi = 0.1 + zeros(1,N); % steering input

for i=2:N
    x(i) = x(i-1) + deltaT*vel(i-1)*cos(theta(i-1));
    y(i) = y(i-1) + deltaT*vel(i-1)*sin(theta(i-1));
    theta(i) = theta(i-1) + (deltaT*vel(i-1)*tan(phi(i-1)))/L;
end
%-------------------------------------------------------------------------


%---------------------EXTENDED KALMAN FILTER------------------------------
% Declare kalman filter components
xhat = zeros(3,N);
xhat_minus = zeros(3,1);
P = zeros(3,3*N);           % 3x3 matrix
P_minus = zeros(3,3);
A = zeros(3,3);             % process Jacobian (transition)
W = zeros(3,3);             % process Jacobian (noise)
K = zeros(3,3);

% Initialize
xhat(1,1) = 0; xhat(2,1) = 0; xhat(3,1) = 0;
P(1,1)=1; P(1,2)=1; P(1,3)=1;
P(2,1)=1; P(2,2)=1; P(2,3)=1;
P(3,1)=1; P(3,2)=1; P(3,3)=1;


for k=2:N
    % Calculate time update equations
    xhat_minus(1) = xhat(1,k-1) + deltaT*vel(k-1)*cos(xhat(3,k-1));
    xhat_minus(2) = xhat(2,k-1) + deltaT*vel(k-1)*sin(xhat(3,k-1));
    xhat_minus(3) = xhat(3,k-1) + (deltaT*vel(k-1)*tan(phi(k-1)))/L;
    
    A = [1, 0, -deltaT*vel(k-1)*sin(xhat(3,k-1)); 0, 1, deltaT*vel(k-1)*cos(xhat(3,k-1)); 0, 0, 1];
    Q = zeros(3,3);
    R = eye(3);
    
    P_minus = A*P(:,3*k-5:3*k-3)*A' + W*Q*W';    % A*(3 columns of P at time k-1)*A' + W*Q*W'
    
    % Calculate measurement update equations
    Jh = [1, 0, 0; 0, 1, 0; 0, 0, 1];
    V = [1, 0, 0; 0, 1, 0; 0, 0, 1];
    
    % Sensor Model => we're assuming the states can be directly measure so
    % our model is of the estimated states
    H = [xhat_minus(1); xhat_minus(2); xhat_minus(3)];
    
    K = P_minus*Jh'*inv(Jh*P_minus*Jh' + V*R*V');
    xhat(:,k) = xhat_minus + K*(z(:,k) - H);
    I = eye(size(K*Jh));
    P(:,3*k-2:3*k) = (I - K*Jh)*P_minus;     % (3 columns of P at time k) = (I-K*H)*P_minus
end
%-------------------------------------------------------------------------

% Plot results
figure; hold on;
plot(x,y);                          % plot x-y (true)
plot(z(1,:),z(2,:),'red');          % plot x-y (measured)
plot(xhat(1,:),xhat(2,:),'green');  % plot x-y (filtered)
legend('sys model','measured','filtered');
xlabel('x [meters]');
ylabel('y [meters]');
axis([0 20 0 20]);

% Plot results
t = 0:1:N-1;
figure; hold on;
plot(t,theta*180/pi);                          % plot x-y (true)
plot(t,z(3,:)*180/pi,'red');          % plot x-y (measured)
plot(t,xhat(3,:)*180/pi,'green');  % plot x-y (filtered)
legend('sys model','measured','filtered');
xlabel('time [secs]');
ylabel('\theta [degs]');