%-------------------------------------------------------------------------
% FILE: kalman_linear.m
% AUTH: B. Green
% DATE: 07/23/06
% DESC: Implements a Kalman filter to eliminate the noise in the 
%       height measurement (simulated) of a free-falling object.
%-------------------------------------------------------------------------

clear all; clc;

%---------------------SIMULATE NOISY MEASUREMENT--------------------------
z = [127.0 115.3, 110.9, 72.4, 50.7, 0.3];
[dummy,samples] = size(z);
deltaT = 1;     % time between samples
%-------------------------------------------------------------------------

%------------------POSITION BASED ON PROCESS MODEL------------------------
y = zeros(1,samples);
v = zeros(1,samples);
y(1) = 125;
v(1) = 0;
a = -9.81;

for i=2:samples
    y(i) = y(i-1) + v(i-1)*deltaT + 0.5*a*deltaT^2;
    v(i) = v(i-1) + a*deltaT;
end
%-------------------------------------------------------------------------


% Declare kalman filter components
xhat = zeros(2,samples);
P = zeros(2,samples*2);     % 2x2 matrix

xhat_minus = zeros(2,1);
P_minus = zeros(2,2);       % 2x2 matrix
K = zeros(2,1);
u = a + zeros(1,samples);   % control vector is acceleration

% Initialize kalman filter components
A = [1 deltaT; 0 1];        % derived from projectile motion equations
B = [deltaT^2/2; deltaT];   % derived from projectile motion equations
H = [1 0];                  % noisy measurement is of height directly
P(1,1)=1; P(1,2)=1;         % if exact starting location is known, then this matrix should be zero
P(2,1)=1; P(2,2)=1;         % if exact starting location is known, then this matrix should be zero
Q = [0 0; 0 0];             % state noise covariance
R = 1;                      % measurement noise covariance
xhat(1,1) = 100;            % initial position estimate is chosen as 100
xhat(2,1) = 0;              % initial velocity estimate is chosen as 0

for k=2:samples
    % Calculate time update equations
    xhat_minus = A*xhat(:,k-1) + B*u(k-1);
    P_minus = A*P(:,2*k-3:2*k-2)*A' + Q;
    
    % Calculate measurement update equations
    K = P_minus*H'*inv(H*P_minus*H' + R);
    xhat(:,k) = xhat_minus + K*(z(k) - H*xhat_minus);
    I = eye(size(K*H));
    P(:,2*k-1:2*k) = (I - K*H)*P_minus;
end

% Plot results
t = 0:1:samples-1;
sysHeight = y;
figure;
plot(t,sysHeight);
hold on;
plot(t,z,'red');
plot(t,xhat(1,:),'green');
legend('h_{sys model}','h_{measured}','h_{kalman}');
xlabel('time');
ylabel('height');

